from typing import Optional

from ontolutils import Thing, namespaces, urirefs
from pydantic import Field


@namespaces(skos="http://www.w3.org/2004/02/skos/core#")
@urirefs(Concept='skos:Concept',
         prefLabel='skos:prefLabel',
         altLabel='skos:altLabel',
         hiddenLabel='skos:hiddenLabel',
         definition='skos:definition',
         note='skos:note',
         scopeNote='skos:scopeNote',
         editorialNote='skos:editorialNote',
         changeNote='skos:changeNote',
         example='skos:example')
class Concept(Thing):
    """Implementation of skos:Concept"""

    prefLabel: Optional[str] = Field(default=None, alias="pref_label")
    altLabel: Optional[str] = Field(default=None, alias="alt_label")
    hiddenLabel: Optional[str] = Field(default=None, alias="hidden_label")
    definition: Optional[str] = Field(default=None)
    note: Optional[str] = Field(default=None)
    scopeNote: Optional[str] = Field(default=None, alias="scope_note")
    editorialNote: Optional[str] = Field(default=None, alias="editorial_note")
    changeNote: Optional[str] = Field(default=None, alias="change_note")
    example: Optional[str] = Field(default=None)



import json
import pathlib
import unittest
import warnings

import pydantic
import rdflib

import ssnolib
from ssnolib import dcat

__this_dir__ = pathlib.Path(__file__).parent
CACHE_DIR = ssnolib.utils.get_cache_dir()


def _delete_test_data():
    for _filename_to_delete in (CACHE_DIR / 'cf-standard-name-table.xml',
                                CACHE_DIR / 'cfsnt.json',
                                CACHE_DIR / 'cf_table.json',
                                CACHE_DIR / 'cfsnt.json',
                                CACHE_DIR / 'test_snt.yaml',
                                CACHE_DIR / 'cf-standard-name-table.xml',):
        _filename_to_delete.unlink(missing_ok=True)


class TestClasses(unittest.TestCase):

    def setUp(self) -> None:
        warnings.filterwarnings("ignore", category=UserWarning)
        _delete_test_data()

    def tearDown(self) -> None:
        _delete_test_data()

    def test_Organization(self):
        orga = ssnolib.Organization(name='My Orga')
        self.assertEqual(orga.name, 'My Orga')
        orga2 = ssnolib.Organization(name='My Orga', mbox='my.orga@orga.org')
        self.assertEqual(orga2.name, 'My Orga')
        self.assertEqual(str(orga2.mbox), 'my.orga@orga.org')
        self.assertNotEqual(orga, orga2)
        self.assertEqual(orga, orga)

    def test_Person(self):
        contact = ssnolib.Person(mbox='johndoe@email.com')
        self.assertEqual(str(contact.mbox), 'johndoe@email.com')

        contact = ssnolib.Person(firstName='John', lastName='Doe')
        self.assertEqual(contact.firstName, 'John')
        self.assertEqual(contact.lastName, 'Doe')

        # check aliases:
        contact_alias = ssnolib.Person(first_name="John", last_name="Doe")
        self.assertEqual(contact_alias.firstName, 'John')
        self.assertEqual(contact_alias.first_name, 'John')
        self.assertEqual(contact_alias.lastName, 'Doe')
        self.assertEqual(contact_alias.last_name, 'Doe')

    def test_ssnolib_Distribution(self):
        distribution = ssnolib.dcat.Distribution(title='XML Table',
                                                 downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
                                                 mediaType='text/csv')
        self.assertEqual(str(distribution.mediaType),
                         "https://www.iana.org/assignments/media-types/text/csv")

        distribution = ssnolib.dcat.Distribution(title='XML Table',
                                                 downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
                                                 mediaType='application/xml')
        self.assertEqual(str(distribution.mediaType),
                         "https://www.iana.org/assignments/media-types/application/xml")
        self.assertEqual(distribution.title, 'XML Table')
        self.assertEqual(str(distribution.downloadURL),
                         'https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml')

        download_filename = distribution.download('cf-standard-name-table.xml')
        self.assertIsInstance(download_filename, pathlib.Path)
        self.assertTrue(download_filename.exists())
        self.assertTrue(download_filename.is_file())
        download_filename.unlink(missing_ok=True)

    def test_standard_name_table(self):
        snt = ssnolib.StandardNameTable(title='CF Standard Name Table v79')
        self.assertEqual(snt.title, 'CF Standard Name Table v79')
        self.assertEqual(str(snt), 'CF Standard Name Table v79')
        self.assertEqual(repr(snt),
                         f'StandardNameTable(id={snt.id}, title=CF Standard Name Table v79, standardNames=[])')

        distribution = ssnolib.dcat.Distribution(title='XML Table',
                                                 downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
                                                 mediaType='application/xml')
        self.assertEqual(distribution.title, 'XML Table')
        self.assertEqual(str(distribution.downloadURL),
                         'https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml')
        snt = ssnolib.StandardNameTable(title='CF Standard Name Table v79',
                                        distribution=[distribution, ])
        self.assertEqual(snt.distribution[0].title, 'XML Table')
        self.assertEqual(str(snt.distribution[0].downloadURL),
                         'https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml')
        table_filename = snt.distribution[0].download(
            dest_filename=CACHE_DIR / 'cf-standard-name-table.xml',
        )
        self.assertIsInstance(table_filename, pathlib.Path)
        self.assertTrue(table_filename.exists())
        self.assertTrue(table_filename.is_file())
        self.assertEqual(table_filename, CACHE_DIR / 'cf-standard-name-table.xml')
        try:
            snt.distribution[0].download(
                dest_filename=CACHE_DIR / 'cf-standard-name-table.xml',
            )
        except FileExistsError:
            pass
        snt.distribution[0].download(
            dest_filename=CACHE_DIR / 'cf-standard-name-table.xml',
            overwrite_existing=True
        )
        snt_from_xml = snt.parse(table_filename, fmt='xml', make_standard_names_lowercase=True)
        self.assertIsInstance(snt_from_xml.standardNames, list)
        for sn in snt_from_xml.standardNames:
            self.assertIsInstance(sn, ssnolib.StandardName)

        snt_from_xml_dict = snt_from_xml.model_dump(exclude_none=True)

        agent = snt_from_xml_dict['qualifiedAttribution']["agent"]
        agent.pop("id")
        self.assertDictEqual(agent,
                             {'mbox': 'support@ceda.ac.uk',
                              'name': 'Centre for Environmental Data Analysis'})

        with self.assertRaises(pydantic.ValidationError):
            # invalid string for title:
            ssnolib.StandardNameTable(title=123)

        snt_from_xml.title = f'CF Standard Name Table {snt_from_xml.version}'
        with open(CACHE_DIR / 'cfsnt.json', 'w', encoding='utf-8') as f:
            f.write(snt_from_xml.model_dump_jsonld(context=None))

        g = rdflib.Graph().parse(CACHE_DIR / 'cfsnt.json', format='json-ld')
        for s, p, o in g.triples((None, None, None)):
            self.assertIsInstance(p, rdflib.URIRef)

        snt.model_dump_jsonld()

    def test_standard_name(self):
        """describe "air_temperature" from
        https://cfconventions.org/Data/cf-standard-names/current/build/cf-standard-name-table.html"""

        with self.assertRaises(pydantic.ValidationError):
            # invalid unit
            ssnolib.StandardName(
                standardName='air_temperature',
                unit=123,
                description='Air temperature is the bulk temperature of the air, not the surface (skin) temperature.', )

        atemp = ssnolib.StandardName(
            standardName='air_temperature',
            unit='K',
            description='Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')

        self.assertEqual(str(atemp), 'air_temperature')
        self.assertEqual(atemp.standardName, 'air_temperature')
        self.assertEqual(atemp.unit, 'http://qudt.org/vocab/unit/K')
        self.assertEqual(atemp.description,
                         'Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')

        self.assertEqual(str(atemp), 'air_temperature')
        self.assertEqual(atemp.standardNameTable, None)

        # to dict:
        atemp_dict = atemp.model_dump(exclude_none=True)
        self.assertIsInstance(atemp_dict, dict)
        self.assertEqual(atemp_dict['standardName'], 'air_temperature')
        self.assertEqual(atemp_dict['unit'], 'http://qudt.org/vocab/unit/K')
        self.assertEqual(atemp_dict['description'],
                         'Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')

        atemp_json = atemp.model_dump_json()
        self.assertIsInstance(atemp_json, str)
        atemp_json_dict = json.loads(atemp_json)
        self.assertIsInstance(atemp_json_dict, dict)
        self.assertEqual(atemp_json_dict['standardName'], 'air_temperature')
        self.assertEqual(atemp_json_dict['unit'], 'http://qudt.org/vocab/unit/K')
        self.assertEqual(atemp_json_dict['description'],
                         'Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')

        # to json-ld:
        jsonld_string = atemp.model_dump_jsonld()

        with open(CACHE_DIR / 'cf_table.json', 'w') as f:
            f.write(jsonld_string)

        g = rdflib.Graph()
        g.parse(data=jsonld_string, format='json-ld')
        self.assertEqual(len(g), 4)
        for s, p, o in g:
            self.assertIsInstance(s, rdflib.BNode)
            self.assertIsInstance(p, rdflib.URIRef)
            self.assertIsInstance(o, str)

        # serialize with rdflib:
        jsonld_dict = json.loads(jsonld_string)
        # self.assertEqual(jsonld_dict['@context']['@import'], CONTEXT)

        self.assertEqual(jsonld_dict['@type'], 'ssno:StandardName')
        self.assertEqual(jsonld_dict['ssno:standardName'], 'air_temperature')
        self.assertEqual(jsonld_dict['ssno:description'],
                         'Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')
        self.assertEqual(jsonld_dict['ssno:unit'], 'http://qudt.org/vocab/unit/K')

        # http://qudt.org/vocab/unit/K

    def test_snt_from_yaml(self):
        snt_yml_filename = __this_dir__ / 'data/test_snt.yaml'
        distribution = dcat.Distribution(title='XML Table',
                                         downloadURL=f'file:///{snt_yml_filename}',
                                         mediaType='application/yaml')
        filename = distribution.download(CACHE_DIR / 'test_snt.yaml')
        self.assertNotEqual(filename, snt_yml_filename)
        self.assertTrue(pathlib.Path(filename).exists())
        self.assertIsInstance(filename, pathlib.Path)
        snt = ssnolib.StandardNameTable(
            title='Yaml Test SNT',
            distribution=[distribution, ]
        )
        snt.parse(snt.distribution[0])



"""Implementation of classes
- dcat:Resource
- dcat:Distribution
- dcat:Dataset
"""
import pathlib
import re
import shutil
from datetime import datetime
from typing import Union, List

import pydantic
from dateutil import parser
from ontolutils import Thing, as_id
from ontolutils import urirefs, namespaces
from pydantic import HttpUrl, FileUrl, field_validator, Field, model_validator

from ssnolib.utils import download_file
from ..prov import Person, Organization, Agent


@namespaces(dcat="http://www.w3.org/ns/dcat#",
            dcterms="http://purl.org/dc/terms/", )
@urirefs(Resource='dcat:Resource',
         title='dcterms:title',
         description='dcterms:description',
         creator='dcterms:creator',
         version='dcat:version',
         identifier='dcterms:identifier')
class Resource(Thing):
    """Pydantic implementation of dcat:Resource

    .. note::

        More than the below parameters are possible but not explicitly defined here.



    Parameters
    ----------
    title: str
        Title of the resource (dcterms:title)
    description: str = None
        Description of the resource (dcterms:description)
    creator: Union[Person, Organization] = None
        Creator of the resource (dcterms:creator)
    version: str = None
        Version of the resource (dcat:version),
        best following semantic versioning (https://semver.org/lang/de/)
    """
    title: str = None  # dcterms:title
    description: str = None  # dcterms:description
    creator: Union[Person, Organization] = None  # dcterms:creator
    version: str = None  # dcat:version
    identifier: Union[str, HttpUrl] = None  # dcterms:identifier

    @model_validator(mode="before")
    def change_id(self):
        """Change the id to the downloadURL"""
        return as_id(self, "identifier")

    @field_validator('creator', mode='before')
    @classmethod
    def _parse_creator(cls, creator):
        # check if creator is a valid person or oragnisation. if both fail, just pass creator data, it will fail later
        is_person = False
        is_organisation = False
        try:
            person = Person.model_validate(creator, strict=True)
            is_person = True
        except pydantic.ValidationError:
            pass
            # not a person
        try:
            organisation = Organization.model_validate(creator, strict=True)
            is_organisation = True
        except pydantic.ValidationError:
            # not an organisation
            pass
        if is_person and is_organisation:
            return creator  # cannot distinguish between person and organisation
        if is_person:
            return person
        if is_organisation:
            return organisation
        return creator

    @field_validator('identifier', mode='before')
    @classmethod
    def _identifier(cls, identifier):
        """parse datetime"""
        if identifier.startswith('http'):
            return str(HttpUrl(identifier))
        return identifier


@namespaces(dcat="http://www.w3.org/ns/dcat#")
@urirefs(Distribution='dcat:Distribution',
         downloadURL='dcat:downloadURL',
         accessURL='dcat:accessURL',
         mediaType='dcat:mediaType',
         byteSize='dcat:byteSize',
         keyword='dcat:keyword')
class Distribution(Resource):
    """Implementation of dcat:Distribution

    .. note::
        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    downloadURL: Union[HttpUrl, FileUrl]
        Download URL of the distribution (dcat:downloadURL)
    mediaType: HttpUrl = None
        Media type of the distribution (dcat:mediaType).
        Should be defined by the [IANA Media Types registry](https://www.iana.org/assignments/media-types/media-types.xhtml)
    byteSize: int = None
        Size of the distribution in bytes (dcat:byteSize)
    keyword: List[str]
        Keywords for the distribution.
    """
    downloadURL: Union[HttpUrl, FileUrl, pathlib.Path] = Field(default=None, alias='download_URL')
    accessURL: Union[HttpUrl, FileUrl, pathlib.Path] = Field(default=None, alias='access_URL')
    mediaType: HttpUrl = Field(default=None, alias='media_type')  # dcat:mediaType
    byteSize: int = Field(default=None, alias='byte_size')  # dcat:byteSize
    keyword: List[str] = None  # dcat:keyword

    def _repr_html_(self):
        """Returns the HTML representation of the class"""
        if self.downloadURL is not None:
            return f"{self.__class__.__name__}({self.downloadURL})"
        return super()._repr_html_()

    def download(self,
                 dest_filename: Union[str, pathlib.Path] = None,
                 exist_ok: bool = False,
                 **kwargs) -> pathlib.Path:
        """Downloads the distribution
        kwargs are passed to the download_file function, which goes to requests.get()"""

        if self.downloadURL is None:
            raise ValueError('The downloadURL is not defined')

        def _parse_file_url(furl):
            """in windows, we might need to strip the leading slash"""
            fname = pathlib.Path(furl)
            if fname.exists():
                return fname
            fname = pathlib.Path(self.downloadURL.path[1:])
            if fname.exists():
                return fname
            raise FileNotFoundError(f'File {self.downloadURL.path} does not exist')

        if self.downloadURL.scheme == 'file':
            if dest_filename is None:
                return _parse_file_url(self.downloadURL.path)
            else:
                return shutil.copy(_parse_file_url(self.downloadURL.path), dest_filename)
        dest_filename = pathlib.Path(dest_filename or self.downloadURL.path.split('/')[-1])
        if dest_filename.exists():
            return dest_filename
        return download_file(self.downloadURL,
                             dest_filename,
                             exist_ok=exist_ok,
                             **kwargs)

    @field_validator('mediaType', mode='before')
    @classmethod
    def _mediaType(cls, mediaType):
        """should be a valid URI, like: https://www.iana.org/assignments/media-types/text/markdown"""
        if isinstance(mediaType, str):
            if mediaType.startswith('http'):
                return HttpUrl(mediaType)
            elif mediaType.startswith('iana:'):
                return HttpUrl("https://www.iana.org/assignments/media-types/" + mediaType.split(":", 1)[-1])
            elif re.match('[a-z].*/[a-z].*', mediaType):
                return HttpUrl("https://www.iana.org/assignments/media-types/" + mediaType)
        return mediaType

    @field_validator('downloadURL', mode='before')
    @classmethod
    def _downloadURL(cls, downloadURL):
        """a pathlib.Path is also allowed but needs to be converted to a URL"""
        if isinstance(downloadURL, pathlib.Path):
            return str(FileUrl(f'file://{downloadURL.resolve().absolute()}'))
        try:
            return str(FileUrl(downloadURL))
        except ValueError:
            return str(HttpUrl(downloadURL))


@namespaces(dcat="http://www.w3.org/ns/dcat#")
@urirefs(DatasetSeries='dcat:DatasetSeries')
class DatasetSeries(Resource):
    """Pydantic implementation of dcat:DatasetSeries"""


@namespaces(dcat="http://www.w3.org/ns/dcat#",
            prov="http://www.w3.org/ns/prov#",
            dcterms="http://purl.org/dc/terms/")
@urirefs(Dataset='dcat:Dataset',
         creator='dcterms:creator',
         distribution='dcat:distribution',
         modified='dcterms:modified',
         landingPage='dcat:landingPage',
         inSeries='dcat:inSeries',
         theme='dcat:theme')
class Dataset(Resource):
    """Pydantic implementation of dcat:Dataset

    .. note::

        More than the below parameters are possible but not explicitly defined here.

    Parameters
    ----------
    title: str
        Title of the resource (dcterms:title)
    description: str = None
        Description of the resource (dcterms:description)
    creator: Agent = None
        Creator of the resource (dcterms:creator)
    version: str = None
        Version of the resource (dcat:version)
    distribution: List[Distribution] = None
        Distribution of the resource (dcat:Distribution)
    landingPage: HttpUrl = None
        Landing page of the resource (dcat:landingPage)
    modified: datetime = None
        Last modified date of the resource (dcterms:modified)
    inSeries: DatasetSeries = None
        The series the dataset belongs to (dcat:inSeries)
    theme: HttpUrl = None
        A main category of the resource. A resource can have multiple themes.
    """
    # http://www.w3.org/ns/prov#Person, see https://www.w3.org/TR/vocab-dcat-3/#ex-adms-identifier
    creator: Agent = None
    distribution: Union[Distribution, List[Distribution]] = None  # dcat:Distribution
    modified: datetime = None  # dcterms:modified
    landingPage: HttpUrl = Field(default=None, alias='landing_page')  # dcat:landingPage
    inSeries: DatasetSeries = Field(default=None, alias='in_series')  # dcat:inSeries
    theme: HttpUrl = Field(default=None)  # dcat:inSeries

    @field_validator('distribution', mode='before')
    @classmethod
    def _distribution(cls, distribution):
        if not isinstance(distribution, list):
            return [distribution]
        return distribution
        # def _parse_dist(_dist):
        #     if isinstance(_dist, str):
        #         return Distribution(id=_dist)
        #     return _dist
        #
        # if isinstance(distribution, list):
        #     return [_parse_dist(_dist) for _dist in distribution]
        # return _parse_dist(distribution)

    @field_validator('modified', mode='before')
    @classmethod
    def _modified(cls, modified):
        """parse datetime"""
        if isinstance(modified, str):
            return parser.parse(modified)
        return modified



import argparse


def main():
    parser = argparse.ArgumentParser(
        description='ssnolib command line interface'
    )

    parser.add_argument('--app',
                        help='Starts the flask web app',
                        action='store_true')
    parser.add_argument('-p', '--port',
                        help='Port to run the server on',
                        type=int,
                        default=5000)
    parser.add_argument('-H', '--host',
                        help='Host to run the server on',
                        type=str,
                        default='127.0.0.1')
    args = parser.parse_args()

    if args.app:
        from ssnolib.ui import app
        app.app.run(debug=False, host=args.host, port=args.port)



import pathlib
import unittest

from ssnolib.utils import gpfqcs

__this_dir__ = pathlib.Path(__file__).parent


class TestUtils(unittest.TestCase):

    def test_gpfqcs(self):
        # Test with the given example
        input_str = "[component] standard_name [in_medium]"
        positions = gpfqcs(input_str)
        self.assertDictEqual(
            {-1: 'component', 1: 'in_medium'},
            positions
        )

        input_str = "[surface][component] standard_name [surface] [medium] [process] [condition]"
        positions = gpfqcs(input_str)
        self.assertDictEqual(
            {-2: 'surface', -1: 'component', 1: 'surface', 2: 'medium', 3: "process", 4: "condition"},
            positions
        )



import json

import pydantic

import ssnolib
from ssnolib import StandardNameTable
from ssnolib import VectorStandardName
from ssnolib.m4i import TextVariable
from ssnolib.qudt.utils import iri2str
from ssnolib.ssno.standard_name_table import ROLE2IRI


class ErrorMessages:
    def __init__(self):
        self.messages = {}

    def add(self, e: pydantic.ValidationError):
        for errors in e.errors():
            if errors['type'] not in self.messages:
                self.messages[errors['type']] = [errors['msg']]
            else:
                self.messages[errors['type']].append(errors['msg'])


def remove_off_after_on(input_list):
    """
    Remove "off" after "on" in the input list
    """
    result = []
    skip_next = False  # Flag to skip the next item if it is an "off" after "on"

    for i in range(len(input_list)):
        if skip_next:
            skip_next = False  # Reset the flag and skip this iteration
            continue

        if input_list[i] == "on":
            result.append("on")
            # Set flag to skip the next element, since it will be "off"
            skip_next = True
        else:
            result.append(input_list[i])

    return result


def parse_authors(firstnames, last_names, had_roles, orcid_ids, mboxes):
    authors = []
    roles = []
    for (first_name, last_name, role, orcid, mbox) in zip(firstnames, last_names, had_roles, orcid_ids, mboxes):
        person_dict = dict(
            firstName=first_name if first_name != '' else None,
            lastName=last_name if last_name != '' else None,
            orcidId=orcid if orcid != '' else None,
            mbox=mbox if "@" in mbox else None
        )
        for ak, av in person_dict.copy().items():
            if av is None:
                person_dict.pop(ak)
        authors.append(ssnolib.Person(**person_dict))
        roles.append(role)
    qualified_attributions = []
    for (role, author) in zip(roles, authors):
        if role:
            qualified_attributions.append(
                ssnolib.Attribution(agent=author, hadRole=ROLE2IRI[role.lower().replace(" ", "")]))
        else:
            qualified_attributions.append(ssnolib.Attribution(agent=author))
    return qualified_attributions


def parse_organizations(names, urls, ror_ids, had_roles, mboxes):
    organizations = []
    roles = []
    for (name, url, ror_id, role, mbox) in zip(names, urls, ror_ids, had_roles, mboxes):
        organization_dict = dict(
            name=name if name != '' else None,
            url=url if url != '' else None,
            mbox=mbox if "@" in mbox else None,
            hasRorId=ror_id if ror_id != '' else None
        )
        for k, v in organization_dict.copy().items():
            if v is None:
                organization_dict.pop(k)
        organizations.append(ssnolib.Organization(**organization_dict))
        roles.append(role)
    qualified_attributions = []
    for (role, organization) in zip(roles, organizations):
        if role:
            qualified_attributions.append(ssnolib.Attribution(agent=organization,
                                                              hadRole=ROLE2IRI[role.lower()]))
        else:
            qualified_attributions.append(ssnolib.Attribution(agent=organization))
    return qualified_attributions


def fetch_form_data(request, warning_messages=None):
    if warning_messages is None:
        warning_messages = {}
    error_messages = ErrorMessages()

    qa_persons = parse_authors(
        request.form.getlist("person.firstName[]"),
        request.form.getlist("person.lastName[]"),
        request.form.getlist("person.hadRole[]"),
        request.form.getlist("person.orcidId[]"),
        request.form.getlist("person.mbox[]"),
    )
    qa_orgas = parse_organizations(
        request.form.getlist("organization.name[]"),
        request.form.getlist("organization.url[]"),
        request.form.getlist("organization.hasRorId[]"),
        request.form.getlist("organization.hadRole[]"),
        request.form.getlist("organization.mbox[]"),
    )
    has_valid_values_raw = request.form.getlist("hasValidValues[]")
    has_variable_description = request.form.getlist("hasVariableDescription[]")
    qualification_description = request.form.getlist("qualification_description[]")
    qualification_name = request.form.getlist("qualification_name[]")
    _vectors = request.form.getlist("is_vector_qualification[]")

    vector = remove_off_after_on(_vectors)

    prepositions = [p.strip() for p in request.form.getlist("preposition[]")]

    # figure out what is before and what is after the AnyStandardName:
    before_list = []
    after_list = []
    is_before_any_stdname = True
    for i, name in enumerate(qualification_name):
        if name == "AnyStandardName":
            is_before_any_stdname = False
        else:
            if is_before_any_stdname:
                before_list.append(name)
            else:
                after_list.append(name)
    before = {}
    if len(before_list) == 1:
        before = {before_list[0]: str(ssnolib.namespace.SSNO.AnyStandardName)}
    elif len(before_list) > 1:
        for i, b in enumerate(before_list[:-1]):
            before[b] = before_list[i + 1]
    after = {}
    if len(after_list) == 1:
        after = {after_list[0]: str(ssnolib.namespace.SSNO.AnyStandardName)}
    elif len(after_list) > 1:
        for i, a in enumerate(after_list[:-1]):
            after[a] = after_list[i + 1]

    qualification_name.remove("AnyStandardName")

    qualifications = []
    for name, vec, preposition, hvvr, descr in zip(qualification_name,
                                                   vector,
                                                   prepositions,
                                                   has_valid_values_raw,
                                                   qualification_description):
        prep = None if preposition == '' else preposition
        qvalues = []
        for v in hvvr.split(","):
            description = has_variable_description.pop(0)
            qvalues.append(TextVariable(hasStringValue=v.strip(), hasVariableDescription=description))
        if vec == 'on':
            qualifications.append(
                ssnolib.VectorQualification(name=name, hasValidValues=qvalues, description=descr,
                                            hasPreposition=prep)
            )
        else:
            qualifications.append(
                ssnolib.Qualification(name=name, hasValidValues=qvalues, description=descr, hasPreposition=prep)
            )

    for i, qname in enumerate(qualification_name):
        if qname in before:
            _before_value = before[qname]
            if _before_value.startswith("http"):
                qualifications[i].before = _before_value
            else:
                for q in qualifications:
                    if q.name == _before_value:
                        qualifications[i].before = q.id
        elif qname in after:
            _after_value = after[qname]
            if _after_value.startswith("http"):
                qualifications[i].after = _after_value
            else:
                for q in qualifications:
                    if q.name == _after_value:
                        qualifications[i].after = q.id
    qa_persons.extend(qa_orgas)

    standard_name_names = request.form.getlist("standard_name.name[]")
    standard_name_units = request.form.getlist("standard_name.unit[]")
    standard_name_descriptions = request.form.getlist("standard_name.description[]")
    is_vector_standard_name = remove_off_after_on(request.form.getlist("is_vector_standard_name[]"))
    standard_name_classes = [VectorStandardName if is_vector == 'on' else ssnolib.ScalarStandardName for is_vector in
                             is_vector_standard_name]
    assert len(standard_name_names) == len(standard_name_units) == len(standard_name_descriptions) == len(
        is_vector_standard_name), "Length of standard names, units, descriptions and vector flags do not match"

    standard_names = []
    for (cls, name, unit, description) in zip(standard_name_classes, standard_name_names, standard_name_units,
                                              standard_name_descriptions):
        try:
            new_sn = cls(
                standardName=name,
                unit=unit,
                description=description
            )
            standard_names.append(new_sn)
        except pydantic.ValidationError as e:
            error_messages.add(e)

    snt = StandardNameTable(
        title=request.form.get("title"),
        version=request.form.get("version"),
        description=request.form.get("description"),
        qualifiedAttribution=qa_persons,
        hasModifier=qualifications,
        standardNames=standard_names
    )
    json_data = json.loads(snt.model_dump_jsonld())

    has_errors = False
    for k, v in error_messages.messages.items():
        if len(v) > 0:
            has_errors = True
            break

    data, warning_messages = snt_to_cache_data(snt, warning_messages)

    return json_data, data, warning_messages, error_messages, has_errors


def snt_to_cache_data(snt: StandardNameTable, warning_messages):
    if not isinstance(snt.qualifiedAttribution, list):
        if snt.qualifiedAttribution is not None:
            qualified_attributions = [snt.qualifiedAttribution]
        else:
            qualified_attributions = []
    else:
        if snt.qualifiedAttribution is not None:
            qualified_attributions = snt.qualifiedAttribution
        else:
            qualified_attributions = []

    persons = []
    organizations = []
    for qualified_attribution in qualified_attributions:
        if isinstance(qualified_attribution.agent, ssnolib.Person):
            persons.append(qualified_attribution.agent.model_dump(exclude_none=True))
            if qualified_attribution.hadRole:
                persons[-1]["hadRole"] = qualified_attribution.hadRole.rsplit("#", 1)[-1]
        elif isinstance(qualified_attribution.agent, ssnolib.Organization):
            organizations.append(qualified_attribution.agent.model_dump(exclude_none=True))
            if qualified_attribution.hadRole:
                organizations[-1]["hadRole"] = qualified_attribution.hadRole.rsplit("#", 1)[-1]
    modifier = snt.hasModifier or []
    qualifications = [m for m in modifier if isinstance(m, (ssnolib.VectorQualification,
                                                            ssnolib.Qualification))]
    transformations = [m for m in modifier if isinstance(m, ssnolib.Transformation)]
    transformations_dict = [t.model_dump(exclude_none=True) for t in transformations]
    for i, transformation in enumerate(transformations):
        for j, character in enumerate(transformation.hasCharacter):
            if character.associatedWith.startswith('_:'):
                for q in qualifications:
                    if q.id == character.associatedWith:
                        transformations_dict[i]["hasCharacter"][j]["associatedWith"] = q.name
                        break

    title = snt.title
    version = snt.version
    description = snt.description
    standard_names = snt.standardNames

    standard_names_data = []
    for sn in standard_names:
        unit_str = iri2str.get(sn.unit, None)
        if unit_str is None:
            warning_messages["UnitParsWarnings"].append(
                f'Could not parse unit "{sn.unit}" of standard name "{sn.standardName}" to human readable unit.'
            )
            unit_str = sn.unit.rsplit('/', 1)[-1]
        standard_names_data.append(
            {'standardName': sn.standardName,
             'unit_str': unit_str,
             'unit': sn.unit,
             'description': sn.description,
             'is_vector': isinstance(sn, VectorStandardName)}
        )

    data = {
        'title': title,
        'version': version,
        'description': description,
        'persons': persons,
        'organizations': organizations,
        'qualifications': qualifications,
        'transformations': transformations_dict,
        'standard_names': standard_names_data
    }
    return data, warning_messages



import ast
import importlib
import pathlib
import sys
import unittest

from ontolutils import get_urirefs

from ssnolib import SSNO

__this_dir__ = pathlib.Path(__file__).parent


class TestSSNO(unittest.TestCase):

    def test_python_classes(self):
        namespace_names = [str(n).split('#', 1)[-1] for n in list(SSNO.__dict__.values())]
        ssnolib_module_folder = __this_dir__ / "../ssnolib/ssno"

        sys.path.insert(0, str(ssnolib_module_folder.resolve().parent))
        module = importlib.import_module("ssno")
        ignore = ["AgentRole"]
        ignore_filenames = ["__init__.py", "plugins.py", "parser.py"]
        self.assertTrue(ssnolib_module_folder.exists())
        for filename in ssnolib_module_folder.glob("*.py"):
            if filename.name not in ignore_filenames:
                with open(filename, "r", encoding="utf-8") as f:
                    node = ast.parse(f.read(), filename=filename)
                classes = [n.name for n in ast.walk(node) if isinstance(n, ast.ClassDef)]
                for cls_name in classes:
                    if cls_name not in ignore:
                        cls = getattr(module, cls_name)
                        for iri in get_urirefs(cls).values():
                            if "ssno:" in iri:
                                self.assertIn(iri.split(":", 1)[-1], namespace_names)
                        self.assertTrue(cls_name in namespace_names, f"Class {cls_name} in {filename} not in namespace")



import pathlib
import unittest


class TestReadme(unittest.TestCase):

    def tearDown(self):
        pathlib.Path('cf79.jsonld').unlink(missing_ok=True)
        pathlib.Path('air_temperature.jsonld').unlink(missing_ok=True)

    def test_code1(self):
        import ssnolib
        from ssnolib.dcat import Distribution

        distribution = Distribution(
            title='XML Table',
            downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
            mediaType='application/xml'
        )
        snt = ssnolib.StandardNameTable(title='CF Standard Name Table (latest version)',
                                        distribution=distribution)


    def test_code2(self):
        import ssnolib
        from ssnolib.dcat import Distribution

        # Create a distribution object (downloadable XML file containing the standard name table)
        distribution = Distribution(title='XML Table',
                                    downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
                                    mediaType='application/xml')

        # Create a standard name table object
        snt = ssnolib.StandardNameTable(title='CF Standard Name Table v79',
                                        distribution=[distribution, ])

        # To describe this standard name table, we can export the JSON-LD file:
        with open('cf79.jsonld', 'w', encoding='utf-8') as f:
            f.write(snt.model_dump_jsonld())

    def test_code3(self):
        import ssnolib

        air_temp = ssnolib.StandardName(standard_name='air_temperature',
                                        unit='K',
                                        description='Air temperature is the bulk temperature of the air, not the surface (skin) temperature.')

        # write to JSON-LD
        with open('air_temperature.jsonld', 'w') as f:
            f.write(air_temp.model_dump_jsonld())



from typing import Optional, Union, List, Tuple, Any

from ontolutils import Thing, namespaces, urirefs
from pydantic import Field
from pydantic import field_validator
from pydantic.functional_validators import WrapValidator
from typing_extensions import Annotated

from .. import StandardNameTable
from ..dcat import Dataset as DcatDataset


def is_internal_hdf5_path(path: str, handler):
    if not path.startswith('/'):
        raise ValueError("HDF5 path must start with '/'")
    return path


def is_hdf5_root_path(path: str, handler):
    if path != '/':
        raise ValueError("HDF5 root path must be '/'")
    return path


HDF5Path = Annotated[str, WrapValidator(is_internal_hdf5_path)]
HDF5RootPath = Annotated[str, WrapValidator(is_internal_hdf5_path)]


@namespaces(hdf5="http://purl.allotrope.org/ontologies/hdf5/1.8#")
@urirefs(Dataset='hdf5:Dataset',
         name='hdf5:name')
class Dataset(Thing):
    """Dataset"""
    name: HDF5Path


@namespaces(hdf5="http://purl.allotrope.org/ontologies/hdf5/1.8#")
@urirefs(Group='hdf5:Group',
         member='hdf5:member',
         name='hdf5:name')
class Group(Thing):
    """hdf5:Group"""
    name: HDF5Path
    member: Any = Field(default=None)

    @field_validator("member", mode="before")
    @classmethod
    def check_member(cls, group_or_dataset):
        if isinstance(group_or_dataset, (List, Tuple)):
            for item in group_or_dataset:
                if not isinstance(item, (Group, Dataset)):
                    raise ValueError("Group member must be of type GroupOrDataset")
            return group_or_dataset
        if not isinstance(group_or_dataset, (Group, Dataset)):
            raise ValueError("Group member must be of type GroupOrDataset")
        return group_or_dataset


@namespaces(hdf5="http://purl.allotrope.org/ontologies/hdf5/1.8#",
            m4i="https://matthiasprobst.github.io/ssno#")
@urirefs(File='hdf5:File',
         rootGroup='hdf5:rootGroup',
         usesStandardNameTable='hdf5:usesStandardNameTable')
class File(Thing):
    """Dataset"""
    rootGroup: Optional[Group] = Field(default=None, alias="root_group")
    usesStandardNameTable: Optional[Union[StandardNameTable, DcatDataset]] = Field(default=None,
                                                                                   alias="uses_standard_name_table")

    @field_validator("rootGroup", mode="before")
    @classmethod
    def _rootGroup(cls, root_group):
        if root_group.name != '/':
            raise ValueError("rootGroup must be of type Group")
        return root_group



import pathlib

import requests.exceptions

import ssnolib
import utils
from ssnolib import dcat, prov

__this_dir__ = pathlib.Path(__file__).parent
CACHE_DIR = ssnolib.utils.get_cache_dir()


class TestDcat(utils.ClassTest):

    def test_Resource(self):
        resource1 = dcat.Resource(
            title='Resource title',
            description='Resource description',
            creator=prov.Person(first_name='John', lastName='Doe'),
            version='1.0',
            identifier='https://example.com/resource'
        )
        self.assertEqual(resource1.id, 'https://example.com/resource')
        self.assertEqual(resource1.title, 'Resource title')
        self.assertEqual(resource1.description, 'Resource description')
        self.assertIsInstance(resource1.creator, prov.Person)
        self.assertEqual(resource1.creator.firstName, 'John')
        self.assertEqual(resource1.creator.lastName, 'Doe')
        self.assertEqual(resource1.version, '1.0')
        self.assertEqual(str(resource1.identifier), 'https://example.com/resource')

    def test_Distribution(self):
        distribution_none_downloadURL = dcat.Distribution(
            id='_:b2',
            title='Distribution title',
            description='Distribution description'
        )
        self.assertEqual(distribution_none_downloadURL.id, '_:b2')
        with self.assertRaises(ValueError):
            distribution_none_downloadURL.download()

        distribution_wrongfile = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            downloadURL='file://path/invalid.txt'
        )
        with self.assertRaises(FileNotFoundError):
            distribution_wrongfile.download()

        distribution1 = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            creator=prov.Person(first_name='John', lastName='Doe'),
            version='1.0',
            identifier='https://example.com/distribution',
            accessURL='https://example.com/distribution',
            downloadURL='https://example.com/distribution/download'
        )
        self.assertEqual(distribution1.id, 'https://example.com/distribution')
        self.assertEqual(distribution1.title, 'Distribution title')
        self.assertEqual(distribution1.description, 'Distribution description')
        self.assertIsInstance(distribution1.creator, prov.Person)
        self.assertEqual(distribution1.creator.firstName, 'John')
        self.assertEqual(distribution1.creator.lastName, 'Doe')
        self.assertEqual(distribution1.version, '1.0')
        self.assertEqual(str(distribution1.identifier), 'https://example.com/distribution')
        self.assertEqual(str(distribution1.accessURL), 'https://example.com/distribution')
        self.assertEqual(str(distribution1.downloadURL), 'https://example.com/distribution/download')

        with self.assertRaises(requests.exceptions.HTTPError):
            distribution1.download(timeout=60)

        piv_dist = dcat.Distribution(
            downloadURL=self.test_jsonld_filename
        )
        filename = piv_dist.download(timeout=60)
        self.assertTrue(filename.exists())
        self.assertEqual(filename.name, 'piv_dataset.jsonld')
        self.assertIsInstance(filename, pathlib.Path)

        local_dist = dcat.Distribution(
            downloadURL=filename
        )
        i = 0
        i_max = 3
        while i < i_max:
            try:
                local_filename = local_dist.download(timeout=60)
                break
            except requests.exceptions.HTTPSConnection as e:
                print(e)
                i += 1
        self.assertTrue(local_filename.exists())
        self.assertEqual(local_filename.name, 'piv_dataset.jsonld')
        self.assertIsInstance(local_filename, pathlib.Path)

        filename.unlink(missing_ok=True)

    def test_Dataset(self):
        dataset1 = dcat.Dataset(
            title='Dataset title',
            description='Dataset description',
            creator=prov.Person(first_name='John', lastName='Doe'),
            version='1.0',
            identifier='https://example.com/dataset',
            distribution=[
                dcat.Distribution(
                    title='Distribution title',
                    description='Distribution description',
                    creator=prov.Person(first_name='John', lastName='Doe'),
                    version='1.0',
                    identifier='https://example.com/distribution',
                    accessURL='https://example.com/distribution',
                    downloadURL='https://example.com/distribution/download'
                )
            ]
        )
        self.assertEqual(dataset1.id, 'https://example.com/dataset')
        self.assertEqual(dataset1.identifier, 'https://example.com/dataset')
        self.assertEqual(dataset1.title, 'Dataset title')
        self.assertEqual(dataset1.description, 'Dataset description')
        self.assertIsInstance(dataset1.creator, prov.Person)
        self.assertEqual(dataset1.creator.firstName, 'John')
        self.assertEqual(dataset1.creator.lastName, 'Doe')
        self.assertEqual(dataset1.version, '1.0')
        self.assertEqual(str(dataset1.identifier), 'https://example.com/dataset')
        self.assertIsInstance(dataset1.distribution[0], dcat.Distribution)
        self.assertEqual(dataset1.distribution[0].title, 'Distribution title')
        self.assertEqual(dataset1.distribution[0].description, 'Distribution description')
        self.assertIsInstance(dataset1.distribution[0].creator, prov.Person)
        self.assertEqual(dataset1.distribution[0].creator.firstName, 'John')
        self.assertEqual(dataset1.distribution[0].creator.lastName, 'Doe')
        self.assertEqual(dataset1.distribution[0].version, '1.0')
        self.assertEqual(str(dataset1.distribution[0].id), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].identifier), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].accessURL), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].downloadURL), 'https://example.com/distribution/download')



import json
import pathlib
import platform
import shutil
import unittest
from datetime import datetime

import h5rdmtoolbox as h5tbx
import pydantic
import requests.exceptions
import yaml
from ontolutils import Thing
from ontolutils.namespacelib.m4i import M4I
from ontolutils.utils.qudt_units import parse_unit

import ssnolib
from ssnolib import Organization, Person, AgentRole
from ssnolib import StandardName, StandardNameTable, Transformation
from ssnolib.dcat import Distribution
from ssnolib.namespace import SSNO
from ssnolib.prov import Attribution
from ssnolib.skos import Concept
from ssnolib.ssno.standard_name import ScalarStandardName, VectorStandardName
from ssnolib.utils import download_file

__this_dir__ = pathlib.Path(__file__).parent

CACHE_DIR = ssnolib.utils.get_cache_dir()

SNT_JSONLD = """{
  "@context": {
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "dcat": "http://www.w3.org/ns/dcat#",
    "dct": "http://purl.org/dc/terms/",
    "prov": "http://www.w3.org/ns/prov#",
    "ssno": "https://matthiasprobst.github.io/ssno#",
    "schema": "https://schema.org/"
  },
  "@type": "ssno:StandardNameTable",
  "dct:title": "OpenCeFaDB Fan Standard Name Table",
    "prov:qualifiedAttribution": [
        {
            "@type": "prov:Attribution",
            "prov:agent": {
                "@type": "prov:Person",
                "foaf:firstName": "Matthias",
                "foaf:lastName": "Probst",
                "m4i:orcidId": "https://orcid.org/0000-0001-8729-0482",
                "@id": "https://orcid.org/0000-0001-8729-0482"
            },
            "prov:hadRole": "http://w3id.org/nfdi4ing/metadata4ing#ContactPerson",
            "@id": "_:Nb2d044e3f4834c249070f5d48f2a2110"
        },
        {
            "@type": "prov:Attribution",
            "prov:agent": {
                "@type": "prov:Person",
                "foaf:mbox": "john@doe.com",
                "foaf:firstName": "John",
                "foaf:lastName": "Doe",
                "schema:affiliation": {
                    "@type": "prov:Organization",
                    "foaf:name": "Awesome Institute",
                    "@id": "_:N9635d7b0338947ebb87c22da2acfc4c5"
                },
                "@id": "_:N74403c73be1c4cd4a724d8dbe8c37c1c"
            },
            "prov:hadRole": "http://w3id.org/nfdi4ing/metadata4ing#Supervisor",
            "@id": "_:Nac1247a4798b4830b19a82feef83bc1b"
        }
    ],
  "ssno:standardNames": [
    {
      "@type": "ssno:StandardName",
      "ssno:standardName": "absolute_pressure",
      "ssno:description": "Pressure is force per unit area. Absolute air pressure is pressure deviation to a total vacuum.",
      "ssni:unit": "Pa",
      "@id": "local:39257b94-d31c-480e-a43c-8ae7f57fae6d"
    },
    {
      "@type": "ssno:StandardName",
      "ssno:standardName": "ambient_static_pressure",
      "ssno:description": "Static air pressure is the amount of pressure exerted by air that is not moving. Ambient static air pressure is the static air pressure of the surrounding air.",
      "ssni:unit": "Pa",
      "@id": "local:0637ec26-310b-4b0a-bf4c-e51d4afccc7d"
    },
    {
      "@type": "ssno:StandardName",
      "ssno:standardName": "ambient_temperature",
      "ssno:description": "Air temperature is the bulk temperature of the air, not the surface (skin) temperature. Ambient air temperature is the temperature of the surrounding air.",
      "ssni:unit": "K",
      "@id": "local:3286d041-a826-4776-9d25-065dae107b55"
    },
    {
      "@type": "ssno:StandardName",
      "ssno:standardName": "auxiliary_fan_rotational_speed",
      "ssno:description": "Number of revolutions of an auxiliary fan.",
      "ssni:unit": "1/s",
      "@id": "local:2a521e9d-4481-4965-9b42-390db2da4c83"
    }
  ]
}"""


class TestSSNOStandardNameTable(unittest.TestCase):

    def tearDown(self):
        pathlib.Path('minimal_snt.jsonld').unlink(missing_ok=True)
        pathlib.Path('snt.json').unlink(missing_ok=True)
        pathlib.Path('snt.yaml').unlink(missing_ok=True)
        pathlib.Path('snt2.yaml').unlink(missing_ok=True)
        pathlib.Path('snt_with_mod.yaml').unlink(missing_ok=True)
        if pathlib.Path(__this_dir__ / 'tmp').exists():
            shutil.rmtree(pathlib.Path(__this_dir__ / 'tmp'))

    def test_add_author(self):
        snt = StandardNameTable.parse(__this_dir__ / 'data/test_snt.yaml')
        KIT = Organization(id="https://ror.org/04t3en479",
                           name="Karlsruhe Institute of Technology, Institute of Thermal Turbomachinery",
                           ror_id="https://ror.org/04t3en479")
        author1 = Person(id="https://orcid.org/0000-0001-8729-0482",
                         orcid_id="https://orcid.org/0000-0001-8729-0482",
                         affiliation=KIT)
        author2 = Person(id="https://orcid.org/0000-0001-9560-500X",
                         orcid_id="https://orcid.org/0000-0001-9560-500X",
                         affiliation=KIT)
        snt.add_author(author1, AgentRole.Contact_Person)
        snt.add_author(author2, AgentRole.Supervisor)
        self.assertEqual(2, len(snt.qualifiedAttribution))
        self.assertEqual(4, len(snt.standardNames))
        self.assertEqual(2, sum([isinstance(sn, VectorStandardName) for sn in snt.standardNames]))
        self.assertEqual(2, sum([isinstance(sn, ScalarStandardName) for sn in snt.standardNames]))

    def test_to_jsonld(self):
        snt = StandardNameTable(title="minimal snt")
        with self.assertRaises(ValueError):
            snt.to_jsonld("minimal_snt.json")
        filename = snt.to_jsonld("minimal_snt.jsonld")
        self.assertEqual(filename.name, "minimal_snt.jsonld")
        self.assertTrue(filename.exists())
        filename.unlink(missing_ok=True)

    def test_multiple_agents(self):
        agent1 = ssnolib.Person(
            firstName="Matthias", lastName="Probst",
            orcidId="https://orcid.org/0000-0001-8729-0482",
            id="https://orcid.org/0000-0001-8729-0482")
        agent2 = ssnolib.Person(
            firstName="John", lastName="Doe")
        snt = ssnolib.StandardNameTable(
            title='SNT from scratch',
            version='v1',
            qualifiedAttribution=[
                Attribution(agent=agent1, hadRole=M4I.ContactPerson),
                Attribution(agent=agent2, hadRole=M4I.Supervisor)
            ]
        )
        self.assertEqual(
            "Matthias",
            snt.qualifiedAttribution[0].agent.firstName
        )
        self.assertEqual(
            "John",
            snt.qualifiedAttribution[1].agent.firstName
        )
        self.assertEqual(
            str(M4I.ContactPerson),
            str(snt.qualifiedAttribution[0].hadRole)
        )
        self.assertEqual(
            str(M4I.Supervisor),
            str(snt.qualifiedAttribution[1].hadRole)
        )

    def test_from_jsonld_file(self):
        snt = StandardNameTable.from_jsonld(__this_dir__ / 'data/simpleSNT.jsonld', limit=1)
        qualifications = [q for q in snt.hasModifier if isinstance(q, ssnolib.Qualification)]
        self.assertEqual(1, len(qualifications))

    def test_standard_name_table_types(self):
        snt = StandardNameTable()
        self.assertIsInstance(snt, StandardNameTable)
        self.assertIsInstance(snt, Concept)
        self.assertIsInstance(snt, Thing)

    def test_describe_table_in_zenodo(self):
        snt = StandardNameTable(
            prefLabel="My fancy standard name table",
            altLabel="My standard name table",
            created=datetime.today().strftime("%Y-%m-%d"),
            subject="https://www.wikidata.org/wiki/Q172145"
        )
        self.assertEqual(snt.prefLabel, "My fancy standard name table")
        self.assertEqual(snt.altLabel, "My standard name table")
        self.assertEqual(snt.created, datetime.today().strftime("%Y-%m-%d"))
        self.assertEqual(snt.subject, "https://www.wikidata.org/wiki/Q172145")

    def test_standard_name_table(self):
        sn1 = StandardName(standard_name='x_velocity',
                           description='x component of velocity',
                           unit='m s-1')
        sn2 = StandardName(standard_name='y_velocity',
                           description='y component of velocity',
                           unit='m s-1')

        snt = StandardNameTable()
        self.assertEqual(0, len(snt.standardNames))
        snt.add_new_standard_name(
            StandardName(standard_name='z_velocity', description='z component of velocity', unit='m s-1')
        )
        self.assertEqual(1, len(snt.standardNames))

        snt = StandardNameTable(standardNames=[sn1, sn2])
        with open('snt.json', 'w') as f:
            f.write(snt.model_dump_jsonld())

        sn_dict = snt.get_standard_name_dict()
        self.assertEqual(2, len(sn_dict))
        self.assertEqual('x_velocity', sn_dict['x_velocity'].standardName)
        self.assertEqual('y_velocity', sn_dict['y_velocity'].standardName)

        snt_loaded = list(StandardNameTable.from_jsonld(data=snt.model_dump_jsonld(), limit=None))
        self.assertEqual(len(snt_loaded), 1)
        self.assertEqual(len(snt_loaded[0].standardNames), 2)
        self.assertEqual(snt_loaded[0].standardNames[0].standardName, 'x_velocity')
        self.assertEqual(snt_loaded[0].standardNames[0].description, 'x component of velocity')
        self.assertEqual(snt_loaded[0].standardNames[0].unit, str(parse_unit('m s-1')))
        self.assertEqual(snt_loaded[0].standardNames[1].standardName, 'y_velocity')
        self.assertEqual(snt_loaded[0].standardNames[1].description, 'y component of velocity')
        self.assertEqual(snt_loaded[0].standardNames[1].unit, str(parse_unit('m s-1')))

        snt_loaded = StandardNameTable.from_jsonld(data=snt.model_dump_jsonld(), limit=1)
        self.assertEqual(len(snt_loaded.standardNames), 2)
        self.assertEqual(snt_loaded.standardNames[0].standardName, 'x_velocity')
        self.assertEqual(snt_loaded.standardNames[0].description, 'x component of velocity')
        self.assertEqual(snt_loaded.standardNames[0].unit, str(parse_unit('m s-1')))
        self.assertEqual(snt_loaded.standardNames[1].standardName, 'y_velocity')
        self.assertEqual(snt_loaded.standardNames[1].description, 'y component of velocity')
        self.assertEqual(snt_loaded.standardNames[1].unit, str(parse_unit('m s-1')))
        pathlib.Path('snt.json').unlink(missing_ok=True)

    def test_frozen_standard_name_dataclass(self):
        sn1 = StandardName(standard_name='x_velocity',
                           description='x component of velocity',
                           unit='m s-1')
        sn2 = StandardName(standard_name='y_velocity',
                           description='y component of velocity',
                           unit='m s-1')
        snt = ssnolib.StandardNameTable(standardNames=[sn1, sn2])
        frz_dc = snt.get_standard_names_as_frozen_dataclass()
        self.assertIsInstance(frz_dc.x_velocity, StandardName)
        self.assertEqual(frz_dc.x_velocity.standardName, "x_velocity")
        self.assertIsInstance(frz_dc.y_velocity, StandardName)
        self.assertEqual(frz_dc.y_velocity.standardName, "y_velocity")

    def test_standard_name_from_jsonld(self):
        sn_jsonld = """{"@id": "local:39257b94-d31c-480e-a43c-8ae7f57fae6d",
   "@type": "https://matthiasprobst.github.io/ssno#StandardName",
   "https://matthiasprobst.github.io/ssno#standardName": "absolute_pressure",
   "https://matthiasprobst.github.io/ssno#unit": "Pa",
   "https://matthiasprobst.github.io/ssno#description": "Pressure is force per unit area. Absolute air pressure is pressure deviation to a total vacuum."}"""
        sn = StandardName.from_jsonld(data=json.loads(sn_jsonld))

    def test_standard_name_table_from_large_jsonld(self):
        snt = ssnolib.parse_table(__this_dir__ / "data/cf.jsonld", fmt='jsonld')
        self.assertEqual("cf-standard-name-table", snt.title)
        self.assertEqual("86", snt.version)
        self.assertEqual(4828, len(snt.standardNames))

    def test_standard_name_alternative_parsing(self):
        snt = ssnolib.parse_table(__this_dir__ / "data/simpleSNT.jsonld", fmt='jsonld')
        self.assertEqual("SNT from scratch", snt.title)
        self.assertEqual("v1", snt.version)
        self.assertEqual(2, len(snt.hasModifier))
        self.assertEqual(3, len(snt.standardNames))
        qualifications = [q for q in snt.hasModifier if isinstance(q, ssnolib.VectorQualification)]
        self.assertEqual(1, len(qualifications))
        self.assertEqual("component", qualifications[0].name)
        self.assertEqual("The component of a vector", qualifications[0].description)
        self.assertEqual(3, len(qualifications[0].hasValidValues))
        # check IDs of standard names
        self.assertEqual("https://example.org/001", snt.standardNames[0].id)
        self.assertEqual("_:002", snt.standardNames[1].id)

    def test_standard_name_table_from_jsonld(self):
        snt_jsonld_filename = pathlib.Path(__this_dir__, 'snt.json')
        with open(snt_jsonld_filename, 'w') as f:
            json.dump(json.loads(SNT_JSONLD), f)
        snt = StandardNameTable.parse(snt_jsonld_filename, fmt='jsonld')

        snt_jsonld_filename.unlink(missing_ok=True)
        self.assertEqual(snt.title, 'OpenCeFaDB Fan Standard Name Table')
        self.assertEqual(4, len(snt.standardNames))
        self.assertEqual(snt.standardNames[0].standardName, 'absolute_pressure')
        self.assertEqual(snt.standardNames[0].standard_name, 'absolute_pressure')
        self.assertEqual(snt.standardNames[0].description,
                         'Pressure is force per unit area. Absolute air pressure is pressure deviation to a total vacuum.')
        self.assertEqual(snt.standardNames[0].unit, 'http://qudt.org/vocab/unit/PA')
        self.assertEqual(snt.standardNames[1].standardName, 'ambient_static_pressure')
        self.assertEqual(snt.standardNames[1].standard_name, 'ambient_static_pressure')
        self.assertEqual(snt.standardNames[2].standardName, 'ambient_temperature')
        self.assertEqual(snt.standardNames[2].unit, 'http://qudt.org/vocab/unit/K')

    def test_standard_name_table_from_yaml(self):
        pathlib.Path('snt.yaml').unlink(missing_ok=True)

        dist = Distribution(downloadURL='https://example.org/snt.yaml',
                            mediaType='application/yaml')
        snt = StandardNameTable()
        with self.assertRaises(requests.exceptions.HTTPError):
            snt.parse(dist)

        snt_yaml_data = {'name': 'SNT',
                         'description': 'A testing SNT',
                         'version': 'abc123invalid',  # v1.0.0
                         'identifier': 'https://example.org/sntIdentifier',
                         'standardNames': {'x_velocity': {'description': 'x component of velocity',
                                                          'unit': 'm s-1'},
                                           'y_velocity': {'description': 'y component of velocity',
                                                          'unit': 'm s-1'}}}
        with open('snt.yaml', 'w') as f:
            yaml.dump(snt_yaml_data, f)
        snt = StandardNameTable.parse('snt.yaml', fmt='yaml')
        self.assertEqual(snt.title, 'SNT')

        snt = StandardNameTable.parse('snt.yaml', fmt=None)
        self.assertEqual(snt.title, 'SNT')

    def test_standard_name_table_from_xml(self):
        cf_contention = 'https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml'
        snt_xml_filename = download_file(cf_contention,
                                         dest_filename='snt.xml',
                                         exist_ok=True)
        self.assertTrue(snt_xml_filename.exists())

        xml_snt = StandardNameTable.parse(snt_xml_filename, fmt=None, make_standard_names_lowercase=True)
        self.assertEqual(
            xml_snt.qualifiedAttribution.agent.mbox,
            'support@ceda.ac.uk')

        snt_xml_filename = download_file(cf_contention,
                                         dest_filename='snt.xml',
                                         exist_ok=True)
        self.assertTrue(snt_xml_filename.exists())

        snt_xml_filename = download_file(cf_contention,
                                         dest_filename='snt.xml',
                                         exist_ok=True)
        snt_xml_filename.unlink(missing_ok=True)
        snt_xml_filename = download_file(cf_contention)

        snt = StandardNameTable.parse(snt_xml_filename, fmt='xml', make_standard_names_lowercase=True)
        self.assertEqual(
            snt.qualifiedAttribution.agent.mbox,
            'support@ceda.ac.uk')
        snt_xml_filename.unlink(missing_ok=True)

        dist = Distribution(
            downloadURL='https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml',
            mediaType='application/xml')

        snt = StandardNameTable.parse(dist, make_standard_names_lowercase=True)
        self.assertEqual(
            snt.qualifiedAttribution.agent.mbox,
            'support@ceda.ac.uk'
        )
        self.assertEqual(snt.title, 'cf-standard-name-table')
        pathlib.Path(f'{snt.title}.xml').unlink(missing_ok=True)

        self.assertFalse(snt.verify_name("x_velocity"))

    def test_standard_name_table_to_yaml(self):
        snt_yaml_filename = pathlib.Path('snt.yaml')
        if not snt_yaml_filename.exists():
            self.test_standard_name_table_from_yaml()
        assert snt_yaml_filename.exists()
        snt = StandardNameTable.parse(snt_yaml_filename, fmt='yaml')
        snt.to_yaml('snt2.yaml', overwrite=True)

        with open('snt.yaml') as f:
            yaml1 = yaml.safe_load(f)
            for k, v in yaml1.copy()['standardNames'].items():
                yaml1['standardNames'][k]['unit'] = str(parse_unit(v['unit']))

        with open('snt2.yaml') as f:
            yaml2 = yaml.safe_load(f)

        self.assertDictEqual(yaml1, yaml2)

    def test_scalar_qualifications(self):
        medium = ssnolib.Qualification(name="medium", description="medium of a quantity",
                                       hasValidValues=["air", "water"],
                                       before=ssnolib.SSNO.AnyStandardName)
        snt = ssnolib.StandardNameTable(
            standard_names=[ssnolib.ScalarStandardName(standard_name="density", description="", unit="kg/m^3")])
        snt.hasModifier = [medium, ]
        self.assertTrue(snt.verify_name("density"))
        self.assertTrue(snt.verify_name("air_density"))

    def test_modifications(self):
        comp = ssnolib.Qualification(name="component",
                                     description="component of the vector",
                                     hasValidValues=['x', 'y', "z"])
        medium = ssnolib.Qualification(name="medium", description="medium", hasPreposition='in')
        medium.after = SSNO.AnyStandardName
        comp.before = SSNO.AnyStandardName

        snt = StandardNameTable(title='SNT with modifications')
        snt.hasModifier = [comp, medium]
        self.assertEqual("[component] standard_name [in medium]", snt.get_qualification_rule_as_string())

        snt.to_yaml('snt_with_mod.yaml', overwrite=True)

        new_snt = StandardNameTable.parse('snt_with_mod.yaml', fmt='yaml')
        self.assertEqual(new_snt.title, 'SNT with modifications')
        self.assertEqual("[component] standard_name [in medium]", new_snt.get_qualification_rule_as_string())

        # circular reference:
        medium.before = comp
        comp.after = medium
        # value_error_raised = False
        # recursion_error_raised = False
        # try:
        #     medium.model_dump_jsonld()
        # except ValueError as evalerr:
        #     value_error_raised = True
        # except RecursionError as recerr:
        #     recursion_error_raised = True
        # self.assertTrue(value_error_raised or recursion_error_raised)

    def test_cf_qualifications(self):
        # Taken from https://cfconventions.org/Data/cf-standard-names/docs/guidelines.html#id2797725
        surface = ssnolib.Qualification(
            name="surface",
            description='A surface is defined as a function of horizontal position. Surfaces which are defined using a coordinate value (e.g. height of 1.5 m) are indicated by a single-valued coordinate variable, not by the standard name. In the standard name, some surfaces are named by single words which are placed at the start: toa (top of atmosphere), tropopause, surface. Other surfaces are named by multi-word phrases put after at: at_adiabatic_condensation_level, at_cloud_top, at_convective_cloud_top, at_cloud_base, at_convective_cloud_base, at_freezing_level, at_ground_level, at_maximum_wind_speed_level, at_sea_floor, at_sea_ice_base, at_sea_level, at_top_of_atmosphere_boundary_layer, at_top_of_atmosphere_model, at_top_of_dry_convection. The surface called "surface" means the lower boundary of the atmosphere. sea_level means mean sea level, which is close to the geoid in sea areas. ground_level means the land surface (beneath the snow and surface water, if any). cloud_base refers to the base of the lowest cloud. cloud_top refers to the top of the highest cloud. Fluxes at the top_of_atmosphere_model differ from TOA fluxes only if the model TOA fluxes make some allowance for the atmosphere above the top of the model; if not, it is usual to give standard names with toa to the fluxes at the top of the model atmosphere.',
            hasValidValues=["toa", "tropopause", "surface"]
        )
        component = ssnolib.Qualification(
            name="component",
            description='The direction of the spatial component of a vector is indicated by one of the words upward, downward, northward, southward, eastward, westward, x, y. The last two indicate directions along the horizontal grid being used when they are not true longitude and latitude (if there is a rotated pole, for instance). If the standard name indicates a tensor quantity, two of these direction words may be included, applying to two of the spatial dimensions Z Y X, in that order. If only one component is indicated for a tensor, it means the flux in the indicated direction of the magnitude of the vector quantity in the plane of the other two spatial dimensions. The names of vertical components of radiative fluxes are prefixed with net_, thus: net_downward and net_upward. This treatment is not applied for any kinds of flux other than radiative. Radiative fluxes from above and below are often measured and calculated separately, the "net" being the difference. Within the atmosphere, radiation from below (not net) is indicated by a prefix of upwelling, and from above with downwelling. For the top of the atmosphere, the prefixes incoming and outgoing are used instead.,',
            hasValidValues=["upward", "downward", "northward", "southward", "eastward", "westward", "x", "y"]
        )
        at_surface = ssnolib.Qualification(
            name="surface",
            description=surface.description,
            hasPreposition='at',
            hasValidValues=["adiabatic_condensation_level", "cloud_top", "convective_cloud_top",
                            "cloud_base",
                            "convective_cloud_base", "freezing_level", "ground_level",
                            "maximum_wind_speed_level",
                            "sea_floor", "sea_ice_base", "sea_level", "top_of_atmosphere_boundary_layer",
                            "top_of_atmosphere_model", "top_of_dry_convection"]
        )
        medium = ssnolib.Qualification(
            name="medium",
            description='A medium indicates the local medium or layer within which an intensive quantity applies: in_air, in_atmosphere_boundary_layer, in_mesosphere, in_sea_ice, in_sea_water, in_soil, in_soil_water, in_stratosphere, in_thermosphere, in_troposphere.',
            hasPreposition='in',
            hasValidValues=['air', 'atmosphere_boundary_layer', "mesosphere", "sea_ice", "sea_water", "soil",
                            "soil_water", "stratosphere", "thermosphere", "troposphere"]
        )
        process = ssnolib.Qualification(
            name="process",
            description='The specification of a physical process by the phrase due_to_process means that the quantity named is a single term in a sum of terms which together compose the general quantity named by omitting the phrase. Possibilites are: due_to_advection, due_to_convection, due_to_deep_convection, due_to_diabatic_processes, due_to_diffusion, due_to_dry_convection, due_to_gravity_wave_drag, due_to_gyre, due_to_isostatic_adjustment, due_to_large_scale_precipitation, due_to_longwave_heating, due_to_moist_convection, due_to_overturning, due_to_shallow_convection, due_to_shortwave_heating, due_to_thermodynamics (referring to sea ice freezing and melting).',
            hasPreposition='due_to',
            hasValidValues=["advection", "convection", "deep_convection", "diabatic_processes",
                            "diffusion", "dry_convection", "gravity_wave_drag", "gyre", "isostatic_adjustment",
                            "large_scale_precipitation", "longwave_heating", "moist_convection", "overturning",
                            "shallow_convection", "shortwave_heating", "thermodynamics"]
        )
        condition = ssnolib.Qualification(
            name="condition",
            description='A phrase assuming_condition indicates that the named quantity is the value which would obtain if all aspects of the system were unaltered except for the assumption of the circumstances specified by the condition. Possibilities are assuming_clear_sky, assuming_deep_snow, assuming_no_snow.',
            hasPreposition='assuming',
            hasValidValues=["clear_sky", "deep_snow", "no_snow"]
        )
        self.assertEqual([v.hasStringValue for v in condition.hasValidValues], ["clear_sky", "deep_snow", "no_snow"])

        # order the qualifications:
        from ssnolib.namespace import SSNO
        surface.before = component
        component.before = SSNO.AnyStandardName
        at_surface.after = SSNO.AnyStandardName
        medium.after = at_surface
        process.after = medium
        condition.after = process

        snt = StandardNameTable(
            title='CF Rebuilt'
        )
        snt.hasModifier = [surface, component, at_surface]

        self.assertEqual(
            "[surface] [component] standard_name [at surface]",
            snt.get_qualification_rule_as_string()
        )

        # snt = StandardNameTable(name='CF Rebuilt')
        snt.hasModifier = [surface, component, at_surface, medium, process, condition]

        self.assertEqual(
            "[surface] [component] standard_name [at surface] [in medium] [due to process] [assuming condition]",
            snt.get_qualification_rule_as_string()
        )

        core_scalar_standard_names = [
            ("air_density", "kg m-3"),
            ("air_pressure", "Pa"),
            ("temperature", "K"),
        ]
        core_vector_standard_names = [
            ("coordinate", "m"),
            ("velocity", "m/s"),
        ]
        with self.assertRaises(pydantic.ValidationError):
            snt.append("standardNames", 1.5)

        # snt.standardNames = StandardName(standardName="density", description="density", unit="kg/m-3")
        for csn in core_scalar_standard_names:
            snt.append("standardNames", ScalarStandardName(standardName=csn[0], description="", unit=csn[1]))
        for csn in core_vector_standard_names:
            snt.append("standardNames", VectorStandardName(standardName=csn[0], description="", unit=csn[1]))

        self.assertTrue(
            snt.verify_name("x_air_density")
        )  # density is not a vector standard name, hence x_ cannot be applied!
        self.assertTrue(
            snt.verify_name("x_coordinate")
        )
        self.assertTrue(
            snt.verify_name("tropopause_coordinate")
        )
        self.assertTrue(snt.verify_name("air_density"))  # equals "air_density"
        self.assertTrue(snt.verify_name("tropopause_air_pressure"))  # using regex
        tropopause_air_pressure = snt.get_standard_name("tropopause_air_pressure")  # using regex

        self.assertEqual(
            tropopause_air_pressure.description,
            snt.get_standard_name("air_pressure").description + surface.description)
        snt.add_new_standard_name(
            StandardName(
                standardName="tropopause_air_pressure",
                description="",
                unit="Pa"
            )
        )

    def test_cf_qualifications2(self):
        snt = StandardNameTable()
        surface = ssnolib.Qualification(
            name="surface",
            description="My surface",
            hasValidValues=["toa", "tropopause", "surface"]
        )
        component = ssnolib.VectorQualification(
            name="component",
            description='My component',
            hasValidValues=["x", "y"]
        )
        at_surface = ssnolib.Qualification(
            name="surface",
            description=surface.description,
            hasPreposition='at',
            hasValidValues=["adiabatic_condensation_level", "cloud_top", "convective_cloud_top",
                            "cloud_base",
                            "convective_cloud_base", "freezing_level", "ground_level",
                            "maximum_wind_speed_level",
                            "sea_floor", "sea_ice_base", "sea_level", "top_of_atmosphere_boundary_layer",
                            "top_of_atmosphere_model", "top_of_dry_convection"]
        )
        component.before = surface
        surface.before = SSNO.AnyStandardName
        at_surface.after = SSNO.AnyStandardName
        snt.hasModifier = [surface, component, at_surface]

        self.assertFalse(snt.verify_name("x_velocity"))

        snt.add_new_standard_name(VectorStandardName(standard_name='velocity', unit='m/s', description='velocity'),
                                  False)
        snt.add_new_standard_name(ScalarStandardName(standard_name='pressure', unit='Pa', description='pressure'),
                                  False)
        # self.assertTrue(snt.verify_name("x_velocity"))
        # self.assertTrue(snt.verify_name("toa_velocity"))
        # self.assertFalse(snt.verify_name("invalid_velocity"))
        self.assertTrue(snt.verify_name("x_toa_velocity_at_sea_floor"))
        self.assertFalse(
            snt.verify_name("x_toa_pressure_at_sea_floor"))  # component is not applicable to pressure (scalar)
        # self.assertTrue(snt.verify_name("x_toa_pressure"))
        # self.assertFalse(snt.verify_name("toa_x_velocity"))

    if platform.system() != "Darwin":  # pandoc could not be installed in CI for macOS... to be solved...
        def test_to_html(self):
            cf_contention = 'https://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml'
            snt_xml_filename = download_file(cf_contention,
                                             dest_filename='snt.xml',
                                             exist_ok=True)
            self.assertTrue(snt_xml_filename.exists())

            xml_snt = StandardNameTable.parse(snt_xml_filename, fmt=None, make_standard_names_lowercase=True)
            xml_snt.description = f"This table is built by from {snt_xml_filename}. This is not complete, but an excerpt!"
            # speeding things up:
            xml_snt.standardNames = xml_snt.standardNames[:10]

            surface = ssnolib.Qualification(
                name="surface",
                description='A surface is defined as a function of horizontal position. Surfaces which are defined using a coordinate value (e.g. height of 1.5 m) are indicated by a single-valued coordinate variable, not by the standard name. In the standard name, some surfaces are named by single words which are placed at the start: toa (top of atmosphere), tropopause, surface. Other surfaces are named by multi-word phrases put after at: at_adiabatic_condensation_level, at_cloud_top, at_convective_cloud_top, at_cloud_base, at_convective_cloud_base, at_freezing_level, at_ground_level, at_maximum_wind_speed_level, at_sea_floor, at_sea_ice_base, at_sea_level, at_top_of_atmosphere_boundary_layer, at_top_of_atmosphere_model, at_top_of_dry_convection. The surface called "surface" means the lower boundary of the atmosphere. sea_level means mean sea level, which is close to the geoid in sea areas. ground_level means the land surface (beneath the snow and surface water, if any). cloud_base refers to the base of the lowest cloud. cloud_top refers to the top of the highest cloud. Fluxes at the top_of_atmosphere_model differ from TOA fluxes only if the model TOA fluxes make some allowance for the atmosphere above the top of the model; if not, it is usual to give standard names with toa to the fluxes at the top of the model atmosphere.',
                hasValidValues=["toa", "tropopause", "surface"]
            )
            component = ssnolib.Qualification(
                name="component",
                description='The direction of the spatial component of a vector is indicated by one of the words upward, downward, northward, southward, eastward, westward, x, y. The last two indicate directions along the horizontal grid being used when they are not true longitude and latitude (if there is a rotated pole, for instance). If the standard name indicates a tensor quantity, two of these direction words may be included, applying to two of the spatial dimensions Z Y X, in that order. If only one component is indicated for a tensor, it means the flux in the indicated direction of the magnitude of the vector quantity in the plane of the other two spatial dimensions. The names of vertical components of radiative fluxes are prefixed with net_, thus: net_downward and net_upward. This treatment is not applied for any kinds of flux other than radiative. Radiative fluxes from above and below are often measured and calculated separately, the "net" being the difference. Within the atmosphere, radiation from below (not net) is indicated by a prefix of upwelling, and from above with downwelling. For the top of the atmosphere, the prefixes incoming and outgoing are used instead.,',
                hasValidValues=["upward", "downward", "northward", "southward", "eastward", "westward", "x", "y"]
            )
            at_surface = ssnolib.Qualification(
                name="surface",
                description=surface.description,
                hasPreposition='at',
                hasValidValues=["adiabatic_condensation_level", "cloud_top", "convective_cloud_top",
                                "cloud_base",
                                "convective_cloud_base", "freezing_level", "ground_level",
                                "maximum_wind_speed_level",
                                "sea_floor", "sea_ice_base", "sea_level", "top_of_atmosphere_boundary_layer",
                                "top_of_atmosphere_model", "top_of_dry_convection"]
            )
            medium = ssnolib.Qualification(
                name="medium",
                description='A medium indicates the local medium or layer within which an intensive quantity applies: in_air, in_atmosphere_boundary_layer, in_mesosphere, in_sea_ice, in_sea_water, in_soil, in_soil_water, in_stratosphere, in_thermosphere, in_troposphere.',
                hasPreposition='in',
                hasValidValues=['air', 'atmosphere_boundary_layer', "mesosphere", "sea_ice", "sea_water", "soil",
                                "soil_water", "stratosphere", "thermosphere", "troposphere"]
            )
            process = ssnolib.Qualification(
                name="process",
                description='The specification of a physical process by the phrase due_to_process means that the quantity named is a single term in a sum of terms which together compose the general quantity named by omitting the phrase. Possibilites are: due_to_advection, due_to_convection, due_to_deep_convection, due_to_diabatic_processes, due_to_diffusion, due_to_dry_convection, due_to_gravity_wave_drag, due_to_gyre, due_to_isostatic_adjustment, due_to_large_scale_precipitation, due_to_longwave_heating, due_to_moist_convection, due_to_overturning, due_to_shallow_convection, due_to_shortwave_heating, due_to_thermodynamics (referring to sea ice freezing and melting).',
                hasPreposition='due_to',
                hasValidValues=["advection", "convection", "deep_convection", "diabatic_processes",
                                "diffusion", "dry_convection", "gravity_wave_drag", "gyre", "isostatic_adjustment",
                                "large_scale_precipitation", "longwave_heating", "moist_convection", "overturning",
                                "shallow_convection", "shortwave_heating", "thermodynamics"]
            )
            condition = ssnolib.Qualification(
                name="condition",
                description='A phrase assuming_condition indicates that the named quantity is the value which would obtain if all aspects of the system were unaltered except for the assumption of the circumstances specified by the condition. Possibilities are assuming_clear_sky, assuming_deep_snow, assuming_no_snow.',
                hasPreposition='assuming',
                hasValidValues=["clear_sky", "deep_snow", "no_snow"]
            )

            # order the qualifications:
            surface.before = component
            component.before = SSNO.AnyStandardName
            at_surface.after = SSNO.AnyStandardName
            medium.after = at_surface
            process.after = medium
            condition.after = process

            change_over_time = Transformation(
                name="change_over_time_in_X",
                description="change in a quantity X over a time-interval, which should be defined by the bounds of the time coordinate.",
                altersUnit="[X]",
                hasCharacter=[ssnolib.Character(character="X", associatedWith=SSNO.AnyStandardName)]
            )
            component_derivative_of_X = Transformation(
                name="C_derivative_of_X",
                description="derivative of X with respect to distance in the component direction, which may be northward, "
                            "southward, eastward, westward, x or y. The last two indicate derivatives along the axes of "
                            "the grid, in the case where they are not true longitude and latitude.",
                altersUnit="[X]/[C]",
                hasCharacter=[ssnolib.Character(character="C", associatedWith=component),
                              ssnolib.Character(character="X", associatedWith=SSNO.AnyStandardName)]
            )
            xml_snt.hasModifier = [
                surface,
                component,
                at_surface,
                medium,
                process,
                condition,
                change_over_time,
                component_derivative_of_X
            ]

            with self.assertRaises(ValueError):
                _ = xml_snt.to_html(folder="tmp", filename="snt.html")

            filename = xml_snt.to_html()
            self.assertTrue(filename.exists())
            self.assertEqual(filename.name, f"{xml_snt.title}.html")
            filename.unlink(missing_ok=True)
            snt_xml_filename.unlink(missing_ok=True)

            # save using a custom file:
            filename = xml_snt.to_html(filename='snt.html')
            self.assertTrue(filename.exists())
            self.assertEqual(filename.name, "snt.html")
            filename.unlink(missing_ok=True)
            snt_xml_filename.unlink(missing_ok=True)

            # save to folder:
            folder = __this_dir__ / 'tmp'
            folder.mkdir(exist_ok=True, parents=True)
            filename = xml_snt.to_html(folder=folder)
            self.assertTrue(filename.exists())
            self.assertEqual(filename.parent.name, "tmp")
            self.assertEqual(filename.name, f"{xml_snt.title}.html")
            filename.unlink(missing_ok=True)
            snt_xml_filename.unlink(missing_ok=True)

    def test_regex(self):
        a = ssnolib.Qualification(
            name="a",
            description='a',
            hasValidValues=["a", "aa", "aaa"]
        )
        b = ssnolib.Qualification(
            name="b",
            description='b',
            hasValidValues=["b", "bb", "bbb"]
        )
        c = ssnolib.Qualification(
            name="c",
            description='c',
            hasValidValues=["c", "cc", "ccc"]
        )
        d = ssnolib.Qualification(
            name="d",
            description='d',
            hasValidValues=["d", "dd", "ddd"]
        )

        # order the qualifications:
        from ssnolib.namespace import SSNO
        a.before = b
        b.before = SSNO.AnyStandardName
        c.after = SSNO.AnyStandardName
        d.after = c

        snt = StandardNameTable(name='CF Rebuilt')
        snt.hasModifier = [a, b, c, d]

        core_standard_names = [
            ("density", "kg m-3", "Density of a fluid or gas."),
            ("pressure", "Pa"),
            ("temperature", "K"),
            ("speed", "m/s")
        ]

        # snt.standardNames = StandardName(standardName="density", description="density", unit="kg/m-3")
        for csn in core_standard_names:
            snt.append("standardNames", StandardName(standardName=csn[0], description="", unit=csn[1]))

        self.assertEqual(
            r"^(?:(a|aa|aaa))?_?(?:(b|bb|bbb))?_?standard_name_?(?:(c|cc|ccc))?_?(?:(d|dd|ddd))?$",
            snt.get_qualification_regex()[0]
        )
        self.assertTrue(snt.verify_name("density"))  # equals "air_density"
        self.assertTrue(snt.verify_name("a_density"))  # using regex
        self.assertTrue(snt.verify_name("aa_density"))  # using regex
        self.assertTrue(snt.verify_name("aaa_density"))  # using regex
        self.assertTrue(snt.verify_name("a_density_c"))  # using regex
        self.assertTrue(snt.verify_name("a_density_d"))  # using regex
        self.assertTrue(snt.verify_name("a_density_c_d"))  # using regex
        self.assertTrue(snt.verify_name("a_b_density_c"))  # using regex
        self.assertTrue(snt.verify_name("a_b_density_c_d"))  # using regex

    def test_transformation(self):
        # taken from https://cfconventions.org/Data/cf-standard-names/docs/guidelines.html#process
        X = ssnolib.Character(character="X", associatedWith=ssnolib.namespace.SSNO.AnyStandardName)
        Y = ssnolib.Character(character="Y", associatedWith=ssnolib.namespace.SSNO.AnyStandardName)
        t = ssnolib.Transformation(
            name="derivative_of_X_wrt_Y",
            altersUnit="[X]/[Y]",
            hasCharacter=[X, Y],
            description="dX/dY (keeping any other independent variables constant, i.e. the partial derivative if appropriate).")
        snt = StandardNameTable(name='CF Rebuilt')
        snt.hasModifier = [t]
        self.assertEqual(t.altersUnit, "[X]/[Y]")
        self.assertEqual(
            t.description,
            "dX/dY (keeping any other independent variables constant, i.e. the partial derivative if appropriate).")
        self.assertEqual(t.name, "derivative_of_X_wrt_Y")
        self.assertEqual(t.name, snt.hasModifier[0].name)

    def test_hdf5_accessor(self):
        # noinspection PyUnresolvedReferences
        from ssnolib import h5accessor
        with h5tbx.File() as h5:
            h5.create_dataset('u', data=4.3, attrs={'standard_name': 'x_velocity'})
            with self.assertRaises(ValueError):
                h5.ssno.enrich_hdf()

            h5.attrs['snt'] = 'https://doi.org/10.5281/zenodo.10428817'
            h5.ssno.enrich_hdf(standard_name_table_attribute='snt')
            self.assertEqual(h5.u.rdf.predicate['standard_name'],
                             'https://matthiasprobst.github.io/ssno#hasStandardName')
            self.assertEqual(h5.rdf.predicate['snt'],
                             'https://matthiasprobst.github.io/ssno#hasStandardNameTable')
